/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid library code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * Activate/Deactivate code for hierarchical RAID Sets.
 */

#include "internal.h"
#include "devmapper.h"

/* Find smallest disk larger than given minimum. */
static uint64_t _smallest(struct lib_context *lc,
			  struct dmraid_set *rs, uint64_t min)
{
	uint64_t ret = ~0, sectors;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	if (min % rs->stride)
		min += rs->stride;

	list_for_each_entry(r, &rs->sets, list) {
		sectors = calc_total_sectors(lc, r);
		if (sectors > min && sectors < ret)
			ret = sectors;
	}

	list_for_each_entry(rd, &rs->devs, list) {
		if (rd->sectors > min && rd->sectors < ret)
			ret = rd->sectors;
	}

	return ret & ~(((uint64_t) rs->stride) >> 1);
}

/*
 * Definitions of mappings.
 */

/* Undefined mapping. */
static int dm_undef(struct lib_context *lc, struct dmraid_set *rs, char **table,
		    const char *type_ascii, int operational)
{
	LOG_ERR(0, "Undefined RAID type %u on %s", rs->type, rs->name);
}

/* "Spare mapping". */
static int dm_spare(struct lib_context *lc, struct dmraid_set *rs, char **table,
		    const char *type_ascii, int operational)
{
	log_warn("spare device %s", rs->name);
	return 1;
}

/* Push path and offset onto a table. */
static void _dm_path_offset(char **table, const char *path, uint64_t offset)
{
	p_str(table, " ");
	p_str(table, path);
	p_str(table, " ");
	p_u64(table, offset);
}

/*
 * Create dm table for linear mapping.
 */
static void _dm_linear(char **table, const char *path, const char *type_ascii,
		       uint64_t start, uint64_t sectors, uint64_t offset)
{
	p_u64(table, start);
	p_str(table, " ");
	p_u64(table, sectors);
	p_str(table, " ");
	p_str(table, type_ascii);
	_dm_path_offset(table, path, offset);
}

static int dm_linear(struct lib_context *lc, struct dmraid_set *rs,	
		     char **table, char const *type_ascii, int operational)
{
	unsigned int segments = 0;
	uint64_t start = 0, sectors = 0;
	struct dmraid_dev *rd;
	struct dmraid_set *r;

	list_for_each_entry(r, &rs->sets, list) {
		char *path;

		if (r->status != s_operational && !operational)
			continue;

		path = mkdm_path(lc, r->name);
		sectors = calc_total_sectors(lc, r);
		_dm_linear(table,
			   r->status == s_operational ?
			   segments++, path : DMRAID_ERROR_PATH,
			   type_ascii, start, sectors, 0);
		dbg_free(path);
		start += sectors;
		if (r->list.next != &rs->sets)
			p_str(table, "\n");
	}

	list_for_each_entry(rd, &rs->devs, list) {
		if (rd->status != s_operational && !operational)
			continue;

		_dm_linear(table,
			   rd->status == s_operational ?
			   segments++, rd->di->path : DMRAID_ERROR_PATH,
			   type_ascii, start, rd->sectors, rd->offset);
		start += rd->sectors;
		if (rd->list.next != &rs->devs)
			p_str(table, "\n");
	}

	return segments ? 1 : 0;
}

/*
 * Create dm table for a partitions mapping.
 *
 * Operates on a partitioned RAID set with
 * 1 RAID device defining a partition mapping.
 */
static int dm_partitions(struct lib_context *lc, struct dmraid_set *rs,
			 char **table, const char *type_ascii, int operational)
{
	return dm_linear(lc, rs, table, get_type(lc, t_linear), 1);
}

/*
 * Create dm table for striped mapping taking
 * different disk sizes and the stride size into acccount.
 */
static void _dm_striped_start(char **table, uint64_t min, uint64_t last_min,
			      unsigned int n, unsigned int stride)
{
	p_u64(table, n * last_min);
	p_str(table, " ");
	p_u64(table, n * (min - last_min));
	p_str(table, " striped ");
	p_u64(table, n);
	p_str(table, " ");
	p_u64(table, stride);
}

static int dm_striped(struct lib_context *lc, struct dmraid_set *rs,
		      char **table, char const *type_ascii, int operational)
{
	unsigned int dsks = count_devs(lc, rs), n, stripes = 0;
	uint64_t min = 0, last_min;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	while (dsks) {
		last_min = min;
		if (!(min = _smallest(lc, rs, min)))
			return 0;
	
		n = 0;
		list_for_each_entry(r, &rs->sets, list) {
			if (calc_total_sectors(lc, r) > last_min)
				n++;
		}
	
		list_for_each_entry(rd, &rs->devs, list) {
			if (rd->sectors > last_min)
				n++;
		}
	
		_dm_striped_start(table, min, last_min, n, rs->stride);
		list_for_each_entry(r, &rs->sets, list) {
			if (calc_total_sectors(lc, r) > last_min) {
				char *path = mkdm_path(lc, r->name);

				_dm_path_offset(table,
						r->status == s_operational ?
						stripes++, path :
						DMRAID_ERROR_PATH, 0);
				dbg_free(path);
				dsks--;
			}
		}
	
		list_for_each_entry(rd, &rs->devs, list) {
			if (rd->sectors > last_min) {
				_dm_path_offset(table,
						rd->status == s_operational ?
						stripes++, rd->di->path :
						DMRAID_ERROR_PATH,
						rd->offset);
				dsks--;
			}
		}
	
		if (dsks)
			p_str(table, "\n");
	}

	return stripes ? 1 : 0;
}

/*
 * Create dm table for mirrored mapping.
 */
#define ONE_MB	2*1024*1024
static int dm_raid1(struct lib_context *lc, struct dmraid_set *rs, char **table,
		    char const *type_ascii, int operational)
{
	uint64_t sectors = 0;
	unsigned int max, mirrors = 0, region_size;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	list_for_each_entry(r, &rs->sets, list)
		mirrors++;

	list_for_each_entry(rd, &rs->devs, list)
		mirrors++;

	if (!(sectors = _smallest(lc, rs, 0)))
		return 0;

	/* Calculate dirty log region size */
	max = sectors / 1024;
	if (max > 64 * ONE_MB)
		max = 64 * ONE_MB;
	for (region_size = 128; region_size < max; region_size <<= 1);
	region_size >>= 1;

	p_str(table, "0 ");
	p_u64(table, sectors);
	p_str(table, " ");
	p_str(table, type_ascii);
	p_str(table, " core 2 ");
	p_u64(table, region_size);
	p_str(table, rs->status == s_broken ? " sync " : " nosync ");
	p_u64(table, mirrors);

	mirrors = 0;
	list_for_each_entry(r, &rs->sets, list) {
		if (r->status == s_operational) {
			char *path = mkdm_path(lc, r->name);

			mirrors++;
			_dm_path_offset(table, path, 0);
			dbg_free(path);
		}
	}

	list_for_each_entry(rd, &rs->devs, list) {
		if (rd->status == s_operational) {
			mirrors++;
			_dm_path_offset(table, rd->di->path, rd->offset);
		}
	}

	if (mirrors == 1) {
		dbg_free(*table);
		*table = NULL;
		return dm_linear(lc, rs, table, get_type(lc, t_linear), 0);
	}

	return mirrors ? 1 : 0;
}

/*
 * Activate/deactivate (sub)sets.
 */

/*
 * Array of handler functions for the various types.
 *
 * Functions must be in the same order as in the type enum !!!
 *
 */
struct type_function {
	int(*f)(struct lib_context *lc, struct dmraid_set *rs, char **table,
		char const *type_ascii, int operational);
} type_functions[] = {
	{ dm_undef },
	{ dm_undef },	/* t_group */
	{ dm_partitions },
	{ dm_spare },
	{ dm_linear },
	{ dm_striped },
	{ dm_raid1 },
	/* RAID types below not supported (yet) */
	{ dm_undef },
	{ dm_undef },
	{ dm_undef },
};

/* Deactivate a single set (one level of a device stack). */
static int deactivate_superset(struct lib_context *lc, struct dmraid_set *rs)
{
	int ret = 1, status = dm_status(lc, rs);

	if (lc_opt(lc, LC_TEST))
		log_print("%s [%sactive]", rs->name, status ? "" : "in");
	else if (status)
		ret = dm_remove(lc, rs);
	else
		log_err("RAID set \"%s\" is not active", rs->name);

	return ret;
}

/* Deactivate a RAID set. */
int deactivate_set(struct lib_context *lc, void *v)
{
	struct dmraid_set *r, *rs = v;

	/*
	 * Deactivate myself if not a group set,
	 * which gets never activated itself.
	 */
	if (rs->type != t_group && !deactivate_superset(lc, rs))
		return 0;

	/* Deactivate any subsets recursively */
	list_for_each_entry(r, &rs->sets, list) {
		if (!deactivate_set(lc, (void *) r))
			return 0;
	}

	return 1;
}

/* Activate a single set. */
static int activate_subset(struct lib_context *lc, struct dmraid_set *rs)
{
	int ret = 0;
	char const *type_ascii;
	char *table = NULL;

	if (rs->type == t_group)
		return 1;

	if (!(type_ascii = get_type(lc, rs->type)))
		LOG_ERR(0, "Failed to recognize device-mapper mapping type");

	/* Call type handler */
	if ((ret = (&type_functions[rs->type])->f(lc, rs, &table, type_ascii, 1))) {
		if (lc_opt(lc, LC_TEST))
			log_print("%s: %s", rs->name, table);
		else if (dm_status(lc, rs)) {
			log_err("%s already active!", rs->name);
			ret = 0;
		} else
			ret = dm_create(lc, rs, table);
	} else
		log_err("no mapping possible for RAID set %s", rs->name);

	free_string(&table);

	return ret;
}

/* Activate a RAID set recursively (eg, RAID1 on top of RAID0). */
int activate_set(struct lib_context *lc, void *v)
{
	int ret = 0;
	struct dmraid_set *r, *rs = v;

	if (rs->type != t_group && dm_status(lc, rs)) {
		log_print("RAID set \"%s\" already active", rs->name);
		return 1;
	}

	/* Recursively walk down the chain of stacked RAID sets */
	list_for_each_entry(r, &rs->sets, list) {
		/* Activate set below this one */
		if (!(ret = activate_set(lc, (void *) r)))
			goto out;
	}

	if (!(ret = activate_subset(lc, rs)))
		deactivate_set(lc, rs);

  out:
	return ret;
}
