/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include "internal.h"
#include "ondisk.h"

/*
 * Register a RAID device format handler.
 */
int register_format(struct lib_context *lc, struct dmraid_format *fmt)
{
	struct format_list *fl;

	if (!(fl = dbg_malloc(sizeof(*fl))))
		return 0;

	fl->fmt = fmt;
	list_add_tail(&fl->list, lc_list(lc, LC_FORMATS));

	return 1;
}

/* Return format name */
const char *get_format_name(struct lib_context *lc, void *fmt)
{
	return ((struct dmraid_format *) fmt)->name;
}

/*
 * Register all format handlers.
 *
 * I use an array because of the growing number...
 */
static struct register_fh {
	int (*func)(struct lib_context *lc);
} register_fh[] = {
#include "register.h"
	{ NULL},
};

void unregister_formats(struct lib_context *lc)
{
	struct list_head *elem, *tmp;
	struct format_list *fl;

	list_for_each_safe(elem, tmp, lc_list(lc, LC_FORMATS)) {
		list_del(elem);
		dbg_free(list_entry(elem, typeof(*fl), list));
	}

	INIT_LIST_HEAD(lc_list(lc, LC_FORMATS));
}

int register_formats(struct lib_context *lc)
{
	int ret = 1;
	struct register_fh *fh;

	for (fh = register_fh; fh->func; fh++) {
		if (!(ret = fh->func(lc)))
			goto bad;
	}

	return 1;

  bad:
	/* Clean up in case of error */
	log_err("failed registering format");
	unregister_formats(lc);

	return ret;
}

/* Allocate private space in format handlers (eg, for on-disk metadata). */
void *alloc_private(struct lib_context *lc, const char *name, size_t size)
{
	void *ret;

	if ((ret = dbg_malloc(size)))
		memset(ret, 0, size);
	else
		log_err("failed allocating %s metadata", name);

	return ret;
}

/* Allocate metadata sector array in format handlers. */
void *alloc_meta_areas(struct lib_context *lc, struct dmraid_dev *rd,
		       const char *name, unsigned int n)
{
	void *ret;
	size_t size = n * sizeof(*rd->meta_areas);

	if ((ret = dbg_malloc(size))) {
		memset(ret, 0, size);
		rd->areas = n;
	} else
		log_err("failed allocating %s metadata areas array", name);

	return ret;
}

/* Simple metadata write function for format handlers */
static int _write_metadata(struct lib_context *lc, const char *handler,
			   struct dmraid_dev *rd, int idx, int erase)
{
	int ret = 0;
	void *p, *tmp;

	if (idx >= rd->areas)
		goto out;

	p = tmp = rd->meta_areas[idx].area;
	if (erase &&
	    !(p = alloc_private(lc, handler, rd->meta_areas[idx].size << 9)))
		goto out;

	ret = write_file(lc, rd->di->path, (void *) p,
			rd->meta_areas[idx].size << 9,
			rd->meta_areas[idx].offset << 9);

	log_dbg("writing metadata to %s, offset %" PRIu64 " sectors, size %zu "
		"sectors returned %d", rd->di->path, rd->meta_areas[idx].offset,
		rd->meta_areas[idx].size, ret); 
	if (p != tmp)
		dbg_free(p);

  out:
	return ret;
}

int write_metadata(struct lib_context *lc, const char *handler,
		   struct dmraid_dev *rd, int idx, int erase)
{
	unsigned int i;

	if (idx > -1)
		return _write_metadata(lc, handler, rd, idx, erase);

	for (i = 0; i < rd->areas; i++) {
		if (!_write_metadata(lc, handler, rd, i, erase))
			return 0;
	}

	return 1;
}

/*
 * Support function metadata format handlers.
 *
 * Check list of RAID sets.
 */
void check_rs(struct lib_context *lc, struct dmraid_set *rs,
	      int (*check)(struct lib_context *lc, struct dmraid_set *r))
{
	struct dmraid_set *r;

	if (list_empty(&rs->sets))
		return;

	list_for_each_entry(r, &rs->sets, list) {
		log_dbg("checking RAID set \"%s\"", r->name);
		check(lc, r);
	}
}

/* Check if format identifier is valid. */
int check_valid_format(struct lib_context *lc, char *name)
{
	struct format_list *fl;

	list_for_each_entry(fl, lc_list(lc, LC_FORMATS), list) {
		if (!strcmp(name, get_format_name(lc, fl->fmt)))
			return 1;
	}

	return 0;
}


/*
 * Update RAID set state based on operational subsets/devices.
 *
 * In case of a RAID set hierachy, check availability of subsets
 * and set superset to broken in case *all* subsets are broken.
 * If at least one is still available, set to inconsistent.
 *
 * In case of lowest level RAID stes, check consistence of devices
 * and make the above decision at the device level.
 */
static void _set_rs_status(struct lib_context *lc, struct dmraid_set *rs,
			   unsigned int i, unsigned int operational)
{
	if (operational == i)
		rs->status = s_operational;
	else if (operational > 0)
		rs->status = s_inconsistent;
	else
		rs->status = s_broken;

	log_dbg("set status of set \"%s\" to %u", rs->name, rs->status);
}

enum status set_rs_status(struct lib_context *lc, struct dmraid_set *rs)
{
	unsigned int i = 0, operational = 0;
	struct dmraid_set *r;
	struct dmraid_dev *rd;

	/* Set status of subsets. */
	if (list_empty(&rs->devs)) {
		list_for_each_entry(r, &rs->sets, list) {
			/* Check subsets to set status of superset. */
			i++;
			if (r->status == s_operational ||
			    r->status == s_inconsistent)
				operational++;
	
		}
	} else {
		/* Check status of devices... */
		list_for_each_entry(rd, &rs->devs, list) {
			i++;
			if (rd->status == s_operational)
				operational++;
		}
	}
	
	_set_rs_status(lc, rs, i, operational);
	return rs->status;
}
