/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include <errno.h>
#include "internal.h"
#include "misc.h"

/* Create directory recusively. */
static int mk_dir_recursive(struct lib_context *lc, const char *dir)
{
        char *orig, *s;

        log_notice("creating directory %s", dir);

        /* Create parent directories */
        orig = s = dbg_strdup((char *) dir);
	while (1) {
		if ((s = strchr(s + 1, '/')))
			*s = '\0';

		if (mkdir(orig, 0777) && errno != EEXIST) {
			log_err("mkdir %s", orig);
			dbg_free(orig);
			return 0;
		}

		if (s)
			*s++ = '/';
		else
			break;
	}

	dbg_free(orig);
        return 1;
}

/* Create directory. */
int mk_dir(struct lib_context *lc, const char *dir)
{
	struct stat info;

	if (stat(dir, &info))
		return mk_dir_recursive(lc, dir);

	if (S_ISDIR(info.st_mode))
		return 1;

	log_err("directory %s not found", dir);
	return 0;
}

/* Read/Write file offset. */
enum rw {
	READ,
	WRITE
};

static int rw_file(struct lib_context *lc, enum rw rw,
		   int flags, char *path, void *buffer,
		   size_t size, loff_t offset)
{
	int fd, ret = 0;

	if ((fd = open(path, flags, lc->mode)) == -1)
		LOG_ERR(0, "Failed opening device \"%s\"\n", path);

	if (offset && lseek64(fd, offset, SEEK_SET) == (loff_t) -1) {
		log_err("failed seeking device \"%s\"", path);
		goto err;
	}

	if ((ret = (rw == WRITE ? write(fd, buffer, size) :
				  read(fd, buffer, size)) != size)) {
		log_err("failed %sing %s \"%s\"",
			rw == WRITE ? "writ" : "read",
			rw == WRITE ? "to" : "from",
			path);
		ret = 0;
		goto err;
	}

	ret = 1;

  err:
	close(fd);
	return ret;
}

int read_file(struct lib_context *lc, char *path, void *buffer,
	      size_t size, loff_t offset)
{
	return rw_file(lc, READ, O_RDONLY, path, buffer, size, offset);
}

int write_file(struct lib_context *lc, char *path, void *buffer,
	       size_t size, loff_t offset)
{
	return rw_file(lc, WRITE, O_CREAT|O_TRUNC|O_WRONLY, path,
		       buffer, size, offset);
}
