/*
 * Copyright (C) 2004  Heinz Mauelshagen, Red Hat GmbH. All rights reserved.
 *
 * This dmraid code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * dmraid is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this dmraid code; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

/*
 * Tool library
 */

#include <dmraid.h>
#include <ctype.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include "commands.h"
#include "toollib.h"

/* [De]activate a RAID set. */
static int _activate_set(struct lib_context *lc, void *rs, int arg)
{
	if ((ACTIVATE & action) ? activate_set(lc, rs) :
				  deactivate_set(lc, rs)) {
		log_notice("%sctivated %s RAID set %s",
			   action & ACTIVATE ? "A": "Dea", get_set_type(lc, rs),
			   get_set_name(lc, rs));
		return 1;
	}

	return 0;
}

/* [De]activate RAID sets. */
int activate_sets(struct lib_context *lc, int arg)
{
	/* Discover partitions to deactivate RAID sets for and work on them. */
	if (DEACTIVATE & action) {
		discover_partitions(lc);
		process_sets(_activate_set, lc, 0, 1);
	}

	process_sets(_activate_set, lc, arg, 0);

	/* Discover partitions to activate RAID sets for and work on them. */
	if (!(DEACTIVATE & action) && !(NO_PARTITIONS & action)) {
		discover_partitions(lc);
		process_sets(_activate_set, lc, 0, 1);
	}

	return 1;
}

/* Build all sets or the ones given. */
void build_sets(struct lib_context *lc, char **sets)
{
	int o = 0;

	do {
		if (!build_set(lc, sets[o]))
			log_err("failed building set");

		if (!sets[o])
			break;
	} while (sets[++o]);
}

/* Convert a character sting to lower case. */
void str_tolower(char *s)
{
	for (; *s; s++)
		*s = tolower(*s);
}

/* Check if selected or all formats shall be used to read the metadata. */
static int push_format(char ***fmt_names, char *name)
{
	unsigned int i;
	char **p = *fmt_names;

	for (i = 0; p && p[i]; )
		i++;

	if (!(p = dbg_realloc(p, (i + 2) * sizeof (*p))))
		return 0;

	p[i] = name;
	p[i+1] = NULL;
	*fmt_names = p;

	return 1;
}

void free_format_names(struct lib_context *lc)
{
	if (format_names) {
		unsigned int i = 0;

		for (i = 0; format_names[i]; i++)
			dbg_free(format_names[i]);

		dbg_free(format_names);
	}
}

int valid_format(struct lib_context *lc, char *fmt,
		 char ***fmt_names, char **invalid)
{
	int ret = 1;
	char *c, *name, *p = fmt;

	while (*p && ret) {
		if ((c = strchr(p, ',')))
			*c = '\0';

		if (!(name = dbg_strdup(p)))
			return 0;

		if (!(ret = check_valid_format(lc, name)))
			*invalid = name;

		if (!push_format(fmt_names, name))
			return 0;

		p += strlen(p);
		if (c) {
			*c = ',';
			p++;
		}
	}

	return ret;
}

void format_error(struct lib_context *lc, char **fmt_names, const char *error)
{
	if (fmt_names) {
		unsigned int i;

		log_print_nnl("No RAID %s with format: ", error);
		for (i = 0; fmt_names[i]; i++) {
			log_print_nnl("%s", fmt_names[i]);
			if (fmt_names[i+1])
				log_print_nnl(",");
		}
		log_print_nnl("\n");
	} else
		log_print("No Software RAID %ss", error);
}
